import time
from hyperopt import fmin, tpe, hp, STATUS_OK, Trials
# import hyperopt
import sys
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import ShuffleSplit
from sklearn.feature_selection import SelectKBest, chi2, mutual_info_classif
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.model_selection import KFold
from xgboost import plot_importance
from matplotlib import pyplot as plt
from xgboost import XGBClassifier
from numpy import loadtxt
from sklearn.utils import shuffle
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.preprocessing import StandardScaler, Normalizer, MaxAbsScaler, RobustScaler, MinMaxScaler, QuantileTransformer
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
from sklearn.utils import shuffle
import numpy as np
import pandas as pd
import datetime as dt


# GPU = False
parallelType = np
scaler = MinMaxScaler()


class CSA_DE:
    def __init__(self, inputX, target, P, B, lb, ub, alpha, sf, cr, L2, parallelType):
        self.parT = parallelType
        self.X = inputX
        self.FVS = inputX.shape[1]
        # self.XwithBias = self.parT.append(self.parT.ones((self.X.shape[0], 1)), self.X, axis=1)
        self.y = target.reshape(-1, 1)
        self.P = P  # P is population size
        self.alpha = alpha  # the number of clones for each antibody
        self.B = B  # eliminate worst %B number of antibodies in the population
        # The number of parameters to be optimized (FVS: Feature Vector Size)
        self.D = self.FVS + 1
        self.lb = lb  # lower bound for parameters
        self.ub = ub  # upper bound for parameters
        self.sf = sf  # scaling factor of DE
        self.cr = cr  # crossover constant of DE
        self.L2 = L2
        self.evaluationNumber = 0
        # self.tmpID = [-1] * self.P
        self.inds = self.parT.arange(self.P) * self.alpha
        self.indsOfAntibodies = self.parT.arange(self.P, dtype=int)

        # self.Antibodies = self.lb + self.parT.random.rand(self.P, self.D) * (self.ub - self.lb)
        self.Antibodies = self.parT.random.uniform(
            self.lb, self.ub, (self.P, self.D))
        self.solution = self.parT.copy(self.Antibodies)
        self.f = self.calculateF(self.Antibodies)

    def findBestAntibody(self):
        index = self.parT.argmax(self.f)
        self.globalMax = self.f[index, 0]
        self.globalParams = self.parT.copy(
            self.Antibodies[index, :].reshape(1, -1))

    def cloning(self):
        self.clones = self.parT.repeat(self.Antibodies, self.alpha, axis=0)

    def localSearchViaDE(self):
        for i in range(self.P * self.alpha):
            # Mutation
            j = i // self.alpha
            inds = self.indsOfAntibodies[self.indsOfAntibodies != j]
            pars = self.parT.random.choice(
                inds, 3, False) * self.alpha + self.parT.random.randint(0, self.alpha, 3)
            arr = self.clones[pars[0], :] + self.sf * \
                (self.clones[pars[2], :] - self.clones[pars[1], :])
            # Crossover
            ar = self.parT.random.rand(self.D)
            param2change = ar <= self.cr
            self.clones[i, param2change] = arr[param2change]
            self.clones[i, param2change] = self.parT.maximum(
                self.lb, self.clones[i, param2change])
            self.clones[i, param2change] = self.parT.minimum(
                self.ub, self.clones[i, param2change])
        self.cloneF = self.calculateF(self.clones)

    def selection(self):
        clone_reshape = self.cloneF.reshape(self.P, self.alpha)
        max_idxs = self.parT.argmax(clone_reshape, axis=1)
        idxs = self.inds + max_idxs
        bestIdxs = (self.cloneF[idxs] > self.f).flatten()
        self.Antibodies[bestIdxs, :] = self.clones[idxs, :][bestIdxs, :]
        self.f[bestIdxs] = self.cloneF[idxs][bestIdxs]

    def receptorEditing(self):
        fIndex = self.parT.argsort(self.f, axis=0)
        N = round(self.P * self.B)
        worstNindex = fIndex[:N, 0]
        new_N_Antibodies = self.parT.random.uniform(
            self.lb, self.ub, (N, self.D))
        new_N_f = self.calculateF(new_N_Antibodies)
        self.Antibodies[worstNindex, :] = new_N_Antibodies
        self.f[worstNindex] = new_N_f

# There is an error in f1_score function, in which fn variable is wrongly calculated. We should add "axis=0".
    def f1_score(self, actual, predicted):
        tp = self.parT.sum(predicted * actual, axis=0)
        fp = self.parT.sum(predicted, axis=0) - tp
        fn = self.parT.sum(actual, axis=0) - tp
        res = self.parT.zeros(tp.shape)
        ind = tp != 0
        precision = tp[ind] / (tp[ind] + fp[ind])
        recall = tp[ind] / (tp[ind] + fn[ind])
        res[ind] = 2 * precision*recall / (precision+recall)
        return res

    def mcc(self, actual, predicted):
        tp = self.parT.sum(predicted * actual, axis=0)
        tn = self.parT.sum((1 - predicted) * (1 - actual), axis=0)
        fp = self.parT.sum(predicted, axis=0) - tp
        fn = self.parT.sum(actual, axis=0) - tp
        mcc = self.parT.zeros(tp.shape)
        pay = (tp*tn - fp*fn)
        payda = (tp+fn)*(tp+fp)*(tn+fn)*(tn+fp)
        ind = payda != 0
        mcc[ind] = pay[ind] / self.parT.sqrt(payda[ind])
        return mcc

    def calculateF(self, foods):
        W = foods[:, 1:]
        b = foods[:, 0]
        A = self.sig(self.X.dot(W.T) + b)
        p = self.parT.where(A >= 0.5, 1, 0)
        f = self.f1_score(self.y, p).reshape(-1, 1)
        self.evaluationNumber += len(f)
        return f

    def sig(self, n):  # Sigmoid function
        return 1 / (1 + self.parT.exp(-n))


class LearnCSA_DE:
    def __init__(self, inputX, target, P, B, lb, ub, alpha, sf, cr, L2, parallelType, evaluationNumber):
        self.parT = parallelType
        self.csaDE = CSA_DE(inputX, target, P, B, lb, ub,
                            alpha, sf, cr, L2, parallelType)
        self.total_numberof_evaluation = evaluationNumber

    def learn(self):
        self.f_values = []
        self.csaDE.findBestAntibody()
        self.f_values.append(self.csaDE.globalMax)

        sayac = 0
        while self.csaDE.evaluationNumber <= self.total_numberof_evaluation:
            self.csaDE.cloning()
            self.csaDE.localSearchViaDE()
            self.csaDE.selection()
            # self.csaDE.receptorEditing()
            sayac += 1
            if sayac % 20 == 0:
                self.csaDE.receptorEditing()
            self.csaDE.findBestAntibody()
            self.f_values.append(self.csaDE.globalMax)

        # self.csaDE.findBestAntibody()
        self.net = self.csaDE.globalParams
        self.globalMax = self.csaDE.globalMax
        # print(f"Evaluation Number: {self.csaDE.evaluationNumber}"); #


class CSA_DE_LR_Model():
    def __init__(self, lb=-32, ub=32, evaluationNumber=120000, P=40, B=0.1, alpha=2, sf=0.1, cr=0.1, L2=0, parallelType=None):
        '''
        lb is lower bound for parameters to be learned
        ub is upper bound for parameters to be learned
        '''
        self.lb = lb
        self.ub = ub
        self.evaluationNumber = evaluationNumber
        self.P = P
        self.B = B
        self.alpha = alpha
        self.sf = sf
        self.cr = cr
        self.L2 = L2
        self.parallelType = parallelType

    def fit(self, trainX, trainY):
        learn = LearnCSA_DE(trainX, trainY, self.P, self.B, self.lb, self.ub, self.alpha,
                            self.sf, self.cr, self.L2, self.parallelType, self.evaluationNumber)
        learn.learn()
        self.net = learn.net

    def sig(self, x):
        return 1 / (1 + self.parallelType.exp(-x))

    def __str__(self):
        return f"lb={self.lb}, ub={self.ub}, evaNumber={self.evaluationNumber}, P={self.P}, alpha={self.alpha}, B={self.B}, sf={self.sf}, cr={self.cr}, L2={self.L2}"

    def f1_score(self, actual, predicted):
        tp = self.parallelType.sum(predicted * actual, axis=0)
        fp = self.parallelType.sum(predicted, axis=0) - tp
        fn = self.parallelType.sum(actual, axis=0) - tp
        f1 = self.parallelType.zeros(tp.shape)
        ind = tp != 0
        precision = tp[ind] / (tp[ind] + fp[ind])
        recall = tp[ind] / (tp[ind] + fn[ind])
        f1[ind] = 2*precision*recall / (precision+recall)
        return f1

    def mcc(self, actual, predicted):
        tp = self.parallelType.sum(predicted * actual, axis=0)
        tn = self.parallelType.sum((1 - predicted) * (1 - actual), axis=0)
        fp = self.parallelType.sum(predicted, axis=0) - tp
        fn = self.parallelType.sum(actual, axis=0) - tp
        mcc = self.parallelType.zeros(tp.shape)
        pay = (tp*tn - fp*fn)
        payda = (tp+fn)*(tp+fp)*(tn+fn)*(tn+fp)
        ind = payda != 0
        mcc[ind] = pay[ind] / self.parallelType.sqrt(payda[ind])
        return mcc

    def score(self, X, y):
        W = self.net[:, 1:]
        b = self.net[:, 0]
        A = self.sig(X.dot(W.T) + b)
        p = self.parallelType.where(A >= 0.5, 1, 0)
        # prediction = self.logsig(self.parallelType.dot(self.parallelType.append(self.parallelType.ones((X.shape[0], 1)), X, axis=1), self.net.T))
        # prediction = self.parallelType.where(prediction >= 0.5, 1, 0)
        y = y.reshape(-1, 1)
        f1 = self.f1_score(y, p)
        mcc = self.mcc(y, p)
        acc = self.parallelType.average(y == p)
        roc_auc = roc_auc_score(y, p)
        confMat = self.getConfusionMatrix(y, p)
        return [acc, f1, mcc, roc_auc, confMat, self.net]

    def getConfusionMatrix(self, actual, predicted):
        # Confusion Matrix by Google
        # TN FN
        # FP TP
        # Confusion Matrix by Wiki and Sklearn
        # TN FP
        # FN TP
        confMat = self.parallelType.zeros((2, 2))
        for i in range(2):
            for j in range(2):
                confMat[i, j] = self.parallelType.sum(
                    predicted[(actual == i)] == j)
        return confMat


datasetNames = ["re_processedStatlog.csv", "clevelanLast.csv"]
basePath = "../../Datasets/DiseasePrediction/"

# datasetName = datasetNames[0]
# data = pd.read_csv(basePath + datasetName)
# X = data.drop(['presence'], axis=1).values
# y = data['presence'].values.astype('int')

datasetName = datasetNames[1]
data = pd.read_csv(basePath + datasetName)
X = data.drop(['target'], axis=1).values
y = data['target'].values.astype('int')

print(f"Sparcity Ratio: {np.sum(X == 0) / (X.shape[0] * X.shape[1])}")
print(f"Imbalance Ratio: {(y == 0).sum()/len(y)}")

iter = 300
repeat = 1
res_dict = {"Id": [], "Order": [], "RunTime": [], "F1": [],
            "ACC": [], "MCC": [], "ROC_AUC": [], "ConfMat": [], "Net": []}
parameter_dict = {
    "Id": [],
    "lb": [],
    "ub": [],
    "evaNumber": [],
    "P": [],
    "alpha": [],
    "B": [],
    "sf": [],
    "cr": [],
    "L2": [],
}


def crossVal(model, X, y, n, id):
    ACCcrossVal = []
    F1crossVal = []
    MCCcrossVal = []
    ROC_AUCcrossVal = []
    RunTimeCrossVal = []
    for index in range(n):
        numberOfSamples = X.shape[0]
        numberOfTest = numberOfSamples // n
        # numberOfTrain = numberOfSamples - numberOfTest
        testStart = index * numberOfTest
        testEnd = (index + 1) * numberOfTest
        testData = X[testStart:testEnd, :]
        trainData = parallelType.append(
            X[:testStart, :], X[testEnd:, :], axis=0)
        trainData = scaler.fit_transform(trainData)
        testData = scaler.transform(testData)
        targetTest = y[testStart:testEnd]
        targetTrain = parallelType.append(y[:testStart], y[testEnd:], axis=0)
        start_time = dt.datetime.now()
        model.fit(trainData, targetTrain)
        run_time = (dt.datetime.now() - start_time).total_seconds()
        [acc, f1, mcc, roc_auc, confMat, net] = model.score(
            testData, targetTest)
        res_dict["Id"].append(id)
        res_dict["Order"].append(index)
        res_dict["ACC"].append(acc)
        res_dict["F1"].append(f1)
        res_dict["MCC"].append(mcc)
        res_dict["ROC_AUC"].append(roc_auc)
        res_dict["ConfMat"].append(confMat.ravel())
        res_dict["RunTime"].append(run_time)
        res_dict["Net"].append(net)
    # print(f"Confusion Matrix Google (TN,FN,FP,TP)\nConfusion Matrix Wiki and Sklearn (TN,FN,FP,TP): {confMat.ravel()}\nRun time: {run_time}")
    # print(acc)
        F1crossVal.append(f1)
        ACCcrossVal.append(acc)
        MCCcrossVal.append(mcc)
        ROC_AUCcrossVal.append(roc_auc)
        RunTimeCrossVal.append(run_time)
# print(f"F1 Cross Val: {F1crossVal}")
# print(f"ACC Cross Val: {ACCcrossVal}")
    res_dict["Id"].append(id)
    res_dict["Order"].append(-1)
    res_dict["ACC"].append(parallelType.average(ACCcrossVal))
    res_dict["F1"].append(parallelType.average(F1crossVal))
    res_dict["MCC"].append(parallelType.average(MCCcrossVal))
    res_dict["ROC_AUC"].append(parallelType.average(ROC_AUCcrossVal))
    print(f"ACC: {parallelType.average(ACCcrossVal)}, F1: {parallelType.average(F1crossVal)}, MCC: {parallelType.average(MCCcrossVal)}, ROC-AUC: {parallelType.average(ROC_AUCcrossVal)}")
    res_dict["ConfMat"].append(-1)
    res_dict["Net"].append(-1)
# res_dict['RunTime'].append(parallelType.sum(listRunTime))
    res_dict["RunTime"].append(-1)
    return F1crossVal


#################################### BAYESSIAN OPTIMIZATION ####################################
maxF1 = 0
counter = 1


def objective(params):
    global maxF1, counter
    lb = params["lb"]
    ub = params["ub"]
    evaluationNumber = 160000
    B = params["B"]
    P = int(params["P"])
    alpha = int(params["alpha"])
    sf = params["sf"]
    cr = params["cr"]
    L2 = 0

    # start_time = dt.datetime.now()
    testResults = []
    print(
        f"lb={lb}, ub={ub}, evaNumber={evaluationNumber}, P={P}, alpha={alpha}, B={B}, sf={sf}, cr={cr}"
    )
    parameter_dict["Id"].append(counter)
    parameter_dict["lb"].append(lb)
    parameter_dict["ub"].append(ub)
    parameter_dict["evaNumber"].append(evaluationNumber)
    parameter_dict["P"].append(P)
    parameter_dict["alpha"].append(alpha)
    parameter_dict["B"].append(B)
    parameter_dict["sf"].append(sf)
    parameter_dict["cr"].append(cr)
    parameter_dict["L2"].append(L2)

    for _ in range(repeat):
        model = CSA_DE_LR_Model(
            lb=lb,
            ub=ub,
            evaluationNumber=evaluationNumber,
            P=P,
            B=B,
            alpha=alpha,
            sf=sf,
            cr=cr,
            L2=L2,
            parallelType=parallelType,
        )
        # start_time = dt.datetime.now()
        listF1 = crossVal(model, X, y, 10, counter)
        # print(f"Total Run time: {dt.datetime.now()-start_time}")
        avgTest = parallelType.average(listF1)

        testResults.append(avgTest)
        print(f"10-fold mean Test Result = {avgTest}")

    maxAvgTest = parallelType.max(testResults)

    if maxAvgTest > maxF1:
        maxF1 = maxAvgTest
        print(f"max F1 = {maxF1}")
    sys.stdout.flush()
    print(counter)
    counter += 1
    return {"loss": 1 - maxAvgTest, "status": STATUS_OK}


def optimize(trial):
    params = {
        "lb": hp.uniform("lb", -64, -16),
        "ub": hp.uniform("ub", 16, 64),
        "B": hp.uniform("B", 0.05, 0.2),
        "alpha": hp.uniform("alpha", 2, 7),
        "P": hp.uniform("P", 10, 80),
        # 'L2':hp.uniform('L2',.0,.001),
        "sf": hp.uniform("sf", 0.01, 2.0),
        "cr": hp.uniform("cr", 0.01, 1.0)
    }

    best = fmin(
        fn=objective,
        space=params,
        algo=tpe.suggest,
        trials=Trials(),
        max_evals=iter
        #   timeout=100,
    )
    return best


start_time = time.time()
trial = Trials()
trial.mybest = None
best = optimize(trial)
print(best)
print(f"best F1: {maxF1}")
print("--- %s seconds for Bayesian optimization ---" %
      (time.time() - start_time))

all_results = pd.DataFrame.from_dict(res_dict)
pars = pd.DataFrame.from_dict(parameter_dict)

all_results.to_csv(
    f"{datasetName}_all_results_f1_bayesian_r1.csv", index=False)
pars.to_csv(f"{datasetName}_parameters_f1_bayesian_r1.csv", index=False)
