import time
from hyperopt import fmin, tpe, hp, STATUS_OK, Trials

# import hyperopt
import sys
from sklearn.ensemble import RandomForestClassifier
# from sklearn.ensemble import GradientBoostingClassifier
from xgboost import XGBClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
# from sklearn.model_selection import KFold
# from xgboost import plot_importance
# from matplotlib import pyplot as plt
# from sklearn.utils import shuffle
from sklearn.preprocessing import MinMaxScaler
# from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
import numpy as np
import pandas as pd
import datetime as dt


scaler = MinMaxScaler()

datasetNames = ["re_processedStatlog.csv", "clevelanLast.csv"]
basePath = "../../Datasets/DiseasePrediction/"

# datasetName = datasetNames[0]
# data = pd.read_csv(basePath + datasetName)
# y = data["presence"].values.astype("int")
# X = data.drop(["presence"], axis=1).values

datasetName = datasetNames[1]
data = pd.read_csv(basePath + datasetName)
y = data["target"].values.astype("int")
X = data.drop(["target"], axis=1).values

print(f"Sparcity Ratio: {np.sum(X == 0) / (X.shape[0] * X.shape[1])}")
print(f"Imbalance Ratio: {(y == 0).sum()/len(y)}")


def acc_score(actual, predicted):
   return np.average(actual == predicted)

def f1_score(actual, predicted):
  tp = np.sum(predicted * actual, axis=0)
  fp = np.sum(predicted, axis=0) - tp
  fn = np.sum(actual, axis=0) - tp
  f1 = np.zeros(tp.shape)
  ind = tp != 0
  precision = tp[ind] / (tp[ind] + fp[ind])
  recall = tp[ind] / (tp[ind] + fn[ind])
  f1[ind] = 2*precision*recall / (precision+recall)
  return f1

def mcc_score(actual, predicted):
  tp = np.sum(predicted * actual, axis=0)
  tn = np.sum((1 - predicted) * (1 - actual), axis=0)
  fp = np.sum(predicted, axis=0) - tp
  fn = np.sum(actual, axis=0) - tp
  mcc = np.zeros(tp.shape)
  pay = (tp*tn - fp*fn)
  payda = (tp+fn)*(tp+fp)*(tn+fn)*(tn+fp)
  ind = payda != 0
  mcc[ind] = pay[ind] / np.sqrt(payda[ind])
  return mcc

def getConfusionMatrix(actual, predicted):
  confMat = np.zeros((2, 2))
  for i in range(2):
    for j in range(2):
      confMat[i, j] = np.sum(predicted[(actual == i)] == j)
  return confMat


iter = 300
repeat = 1

res_dict = {
    "Id": [],
    "Order": [],
    "RunTime": [],
    "F1": [],
    "ACC": [],
    "MCC": [],
    "ROC_AUC": [],
    "ConfMat": []
}
parameter_dict = {
    'params' : []
}


def crossVal(model, X, y, n, id):
    ACCcrossVal = []
    F1crossVal = []
    MCCcrossVal = []
    ROC_AUCcrossVal = []
    RunTimeCrossVal = []
    for index in range(n):
        numberOfSamples = X.shape[0]
        numberOfTest = numberOfSamples // n
        # numberOfTrain = numberOfSamples - numberOfTest
        testStart = index * numberOfTest
        testEnd = (index + 1) * numberOfTest
        testData = X[testStart:testEnd, :]
        trainData = np.append(X[:testStart, :], X[testEnd:, :], axis=0)
        trainData = scaler.fit_transform(trainData)
        testData = scaler.transform(testData)
        targetTest = y[testStart:testEnd]
        targetTrain = np.append(y[:testStart], y[testEnd:], axis=0)
        start_time = dt.datetime.now()
        model.fit(trainData, targetTrain)
        run_time = (dt.datetime.now() - start_time).total_seconds()
        acc = acc_score(targetTest, model.predict(testData))
        f1 = f1_score(targetTest, model.predict(testData))
        mcc = mcc_score(targetTest, model.predict(testData))
        roc_auc = roc_auc_score(targetTest, model.predict(testData))
        confMat = getConfusionMatrix(targetTest, model.predict(testData))

        res_dict["Id"].append(id)
        res_dict["Order"].append(index)
        res_dict["ACC"].append(acc)
        res_dict["F1"].append(f1)
        res_dict["MCC"].append(mcc)
        res_dict["ROC_AUC"].append(roc_auc)
        res_dict["ConfMat"].append(confMat.ravel())
        res_dict["RunTime"].append(run_time)
        # print(f"Confusion Matrix Google (TN,FN,FP,TP)\nConfusion Matrix Wiki and Sklearn (TN,FN,FP,TP): {confMat.ravel()}\nRun time: {run_time}")
        # print(acc)
        F1crossVal.append(f1)
        ACCcrossVal.append(acc)
        MCCcrossVal.append(mcc)
        ROC_AUCcrossVal.append(roc_auc)
        RunTimeCrossVal.append(run_time)
    # print(f"F1 Cross Val: {F1crossVal}")
    # print(f"ACC Cross Val: {ACCcrossVal}")
    res_dict["Id"].append(id)
    res_dict["Order"].append(-1)
    res_dict["ACC"].append(np.average(ACCcrossVal))
    res_dict["F1"].append(np.average(F1crossVal))
    res_dict["MCC"].append(np.average(MCCcrossVal))
    res_dict["ROC_AUC"].append(np.average(ROC_AUCcrossVal))
    print(f"ACC: {np.average(ACCcrossVal)}, F1: {np.average(F1crossVal)}, MCC: {np.average(MCCcrossVal)}, ROC-AUC: {np.average(ROC_AUCcrossVal)}")
    res_dict["ConfMat"].append(-1)
    # res_dict['RunTime'].append(parallelType.sum(listRunTime))
    res_dict["RunTime"].append(-1)
    return ACCcrossVal


#################################### BAYESSIAN OPTIMIZATION ####################################
maxACC = 0
counter = 1


def objective(params):
    global maxACC, counter

    # start_time = dt.datetime.now()
    testResults = []
    print(params)
    
    parameter_dict["params"].append(params)
    

    for i in range(repeat):

        # model = LinearDiscriminantAnalysis(solver="lsqr", shrinkage=params['shrinkage'])
        # model = LogisticRegression(C=params["C"])
        # model = MLPClassifier(learning_rate_init=params["learning_rate_init"], batch_size=params["batch_size"], max_iter=params["max_iter"], hidden_layer_sizes=params["hidden_layer_sizes"])
        # model = RandomForestClassifier(n_estimators=params["n_estimators"])
        # model = XGBClassifier(learning_rate=params["learning_rate"], max_depth=params["max_depth"])
        # model = SVC(C=params["C"], probability=True)
        model = DecisionTreeClassifier(min_samples_split=params["min_samples_split"], min_samples_leaf=params["min_samples_leaf"])

        listACC = crossVal(model, X, y, 10, counter)
        avgTest = np.average(listACC)

        testResults.append(avgTest)
        print(f"10-fold mean Test Result = {avgTest}")

    maxAvgTest = np.max(testResults)

    if maxAvgTest > maxACC:
        maxACC = maxAvgTest
        print(f"max ACC = {maxACC}")
    sys.stdout.flush()
    print(counter)
    counter += 1
    return {"loss": 1 - maxAvgTest, "status": STATUS_OK}


def optimize(trial):

    # params = {'shrinkage': hp.uniform('shrinkage', 0, 1)} #LinearDiscriminantAnalysis
    # params = {'C': hp.uniform('C', 10e-4, 10e+4)} # LogisticRegression
    # params = {
    #    'learning_rate_init': hp.uniform('learning_rate_init', 10e-8, 10e-1),
    #    'batch_size': hp.randint('batch_size', 1, 1024),
    #    'max_iter': hp.randint('max_iter', 1, 50),
    #    'hidden_layer_sizes': hp.randint('hidden_layer_sizes',  2, 40)
    # } # MLP
    # params = {'n_estimators': hp.randint('n_estimators', 1, 200)} # RandomForest
    # params = {'learning_rate': hp.uniform('learning_rate', 0.1, 1), 'max_depth': hp.randint('max_depth', 1, 40)} # XGBClassifier
    # params = {'C': hp.uniform('C', 0.001, 1)} # SVC
    params = {'min_samples_split': hp.randint('min_samples_split', 2, 100), 'min_samples_leaf': hp.randint('min_samples_leaf', 1, 100)} # DecisionTreeClassifier

    best = fmin(
        fn=objective,
        space=params,
        algo=tpe.suggest,
        trials=Trials(),
        max_evals=iter
        #   timeout=100,
    )
    return best


start_time = time.time()
trial = Trials()
trial.mybest = None
best = optimize(trial)
print(best)
print(f"best ACC: {maxACC}")
print("--- %s seconds for Bayesian optimization ---" %(time.time() - start_time))

all_results = pd.DataFrame.from_dict(res_dict)
pars = pd.DataFrame.from_dict(parameter_dict)


all_results.to_csv(f"MLmethods_Results/{datasetName}_all_results_DecisionTreeClassifier_bayesian.csv", index=False)
pars.to_csv(f"MLmethods_Results/{datasetName}_parameters_DecisionTreeClassifier_bayesian.csv", index=False)
